#!/usr/bin/python
'''
GUI for openFOAM

This program launches a guided user interface for openFOAM

Updates:
The solver are implemented as modules
The general structure is changed

Requirements: openFOAM 2.2.2 with paraFOAM running on UBUNTU 
Made by: Henrik Kaald Melb\o
NTNU, Trondheim, Norway
Department Of Chemical Engineering
Contact: henrikkaald@gmail.com
Version: 3.0
Date: 14. Mar. 2014
'''
from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt
from PyQt4.QtGui import *
from intr import UI
from util import TextEditer, Wiew, Loader, replace, textparse, fvSo
from solv import simpleFoam, icoFoam, pisoFoam, scalarTransportFoam
from turb import kEpsilon
import sys, os, time, subprocess, re, shutil
from tempfile import mkstemp
from shutil import move
from os import remove, close
import imp

###################################################
#Initiating Main GUI                              #
###################################################
class Main(QDialog, UI.Ui_openFOAM):
    def __init__(self, parent = None):
        super(Main, self).__init__(parent)
        # Set up the user interface from Designer.
        self.setupUi(self)
        # Connect up the buttons and comboboxes
        self.pushButton.clicked.connect(self.selectFile) #Load mesh button
        self.pushButton_3.clicked.connect(self.load) #Load case button        
        self.pushButton_2.clicked.connect(self.new) #New case button
        self.pushButton_5.clicked.connect(self.fvSc) #Edit fvSchemes 
        self.pushButton_7.clicked.connect(self.fvSo) #Edit fvSolution 
        self.pushButton_6.clicked.connect(self.terminal) #Terminal command
        self.pushButton_4.setEnabled(False)
        self.textEdit.setEnabled(False)
        self.label_14.setEnabled(False)
        self.comboBox_4.setEnabled(False)
        self.pushButton_4.clicked.connect(self.para)
        self.connect(self.comboBox, QtCore.SIGNAL('activated(QString)'), self.combo) # Solvers
        self.connect(self.comboBox_4, QtCore.SIGNAL('activated(QString)'), self.turbComb) # Turbulence
        applyButton = self.buttonBox.button(QtGui.QDialogButtonBox.Apply)
        QtCore.QObject.connect(applyButton, QtCore.SIGNAL("clicked()"), self.accept)
        cancelButton = self.buttonBox.button(QtGui.QDialogButtonBox.Cancel)
        self.checkBox_2.stateChanged.connect(self.nomesh)
        self.textEdit_9.setReadOnly(True)
        self.textEdit_11.setReadOnly(True)
        self.checkBox.stateChanged.connect(self.turbulence) #Turbulence
        self.Schemes.setTabEnabled(6, 0)
        self.Schemes.removeTab(3)

        #Get solvers from solv folder
        cwd = os.getcwd()
        path = cwd+"/solv"
        files = os.listdir(path)
        solvers = set()
        for strings in files:
            pices = strings.rsplit('.',1)
            solvers.add(pices[0])
        solvers.discard("__init__")
        solvers.discard("solver")
        for solver in solvers:
            self.comboBox.addItem(solver)

    #Creates a new case folder
    def new(self):
        global case
        global pwd
        case = QFileDialog.getSaveFileName(self, "New case")
        case = str(case)
        if not case:
            del case
            return
        pwd = os.path.dirname(os.path.realpath(__file__))
        case = str(case)
        os.system("cp -r "+pwd+"/root "+case)
        self.pushButton_4.setEnabled(True)

    #Loads in an older case folder
    def load(self):
        global case
        global pwd
        global mesh
        case = QFileDialog.getExistingDirectory()
        if not case:
            del case
            return
        case = str(case)
        pwd = os.path.dirname(os.path.realpath(__file__))
        os.system("cp -r "+pwd+"/root/templates "+case)
        if os.path.isdir(case+"/constant/polyMesh"):
            self.findBound()
        self.pushButton_4.setEnabled(True)
        mesh = True

    #Get name of program to run, from combobox
    def combo(self, text): #Run
        if text == "None":
            self.textEdit_11.setText("No solver selected")
            try:
                s
            except:
                return
            del s
            return
        global program, properties
        program = str(text)
        exec "s = "+program+"."+program+"()"
        self.textEdit_11.setText(s.description)
        properties = s.properties
        #New solver can lead to a new set of boundaries 
        try:
            mesh
        except:
            return
        self.findBound()

    #Turbulence Combobox
    def turbComb(self, text):
        if text == "None":
            self.Schemes.removeTab(7)
            try:
                s
            except:
                return
            del s
            return
        global turbprop
        program = str(text)
        exec "s = "+program+"."+program+"()"
        turbprop = s.properties
        self.initturbulence()
 
    #Initializing the final preparations for running the solver
    #Remaining checks and finializations
    def accept(self):
        #Read, search and replace parameters
        global delta
        global write
        try:
            case
        except:
            QtGui.QMessageBox.question(self, 'An error has occured', "No case defined", QtGui.QMessageBox.Ok)
            return
        if not self.validate():
            return
        try:
            program
        except:
            QtGui.QMessageBox.question(self, 'An error has occured', "No solver defined", QtGui.QMessageBox.Ok)
            return
        if program == "None":
            QtGui.QMessageBox.question(self, 'An error has occured', "No solver defined", QtGui.QMessageBox.Ok)
            return
        if not os.path.isdir(case+"/constant/polyMesh"):
            QtGui.QMessageBox.question(self, 'An error has occured', "No mesh defined", QtGui.QMessageBox.Ok)
            return
        if not self.condType():
            QtGui.QMessageBox.question(self, 'An error has occured', "Ill defined boundary type", QtGui.QMessageBox.Ok)
            return
        #Get controlDict
        self.control()
        #If turbulence
        try:
            self.turb
        except:
            self.turb = False
        if self.turb == True:
            self.setturbulence()
        #Write to viscosity
        self.visc()
        #Makes boundries reversable
        if not os.path.exists(case+"/"+str(start)):
            os.makedirs(case+"/"+str(start))
        for f in properties:
            shutil.copy2(case+"/templates/"+f+"_tmp", case+"/"+str(start)+"/"+f)
        #Check and set boundaries
        self.cond()
        #Set solver and run (program is run in the loader)
        load = Loader.Loader(case, run, delta, write, program, pwd, start)
        if load.Run() == 1:
            w = Wiew.Wiew(case, pwd)
            w.exec_()

    #Loads in mesh
    def selectFile(self): # Load and run mesh
        try:
            self.meshstate
        except:
            self.meshState = False
        if self.meshState == True:
            return
        global mesh
        global path
        try:
            case
        except:
            QtGui.QMessageBox.question(self, 'An error has occured', "No case defined", QtGui.QMessageBox.Ok)
            return
        path = QFileDialog.getOpenFileName(self, "Load Mesh", case , "*.unv *.ans *.msh *.neu *.geo")
        if not path:
            self.lineEdit.setText("Mesh")
            del path
            return
        self.lineEdit.setText(path)
        path = str(path)
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        #Copy boundarys
        suf = path.rsplit('.',1)
        os.system("cp " + path +" "+ case)
        mesh = path.rsplit('/',1)
        self.defBound(mesh, suf)
        QApplication.restoreOverrideCursor()
        self.findBound()
        #Create mesh, write info to "MehsInfo"

    #Changes the mesh state 
    def nomesh(self, state):
        if state == QtCore.Qt.Checked:
            self.meshState = True

    #Allows for manual input in fvSchemes
    def fvSc(self):
        try:
            case
        except:
            QtGui.QMessageBox.question(self, 'An error has occured', "No case defined", QtGui.QMessageBox.Ok)
            return
        T = TextEditer.TextEditer(case)
        T.exec_()

    def fvSo(self):
        try:
            case
        except:
            QtGui.QMessageBox.question(self, 'An error has occured', "No case defined", QtGui.QMessageBox.Ok)
            return
        T = fvSo.fvSo(case)
        T.exec_()

    #Checks whats kind of mesh is supplied, and does the necessary conversion
    def defBound(self, mesh, suf):
        gr = open(pwd+"/outp/MeshInfo","w")
        if suf[-1] == "unv":
            d = subprocess.Popen(["ideasUnvToFoam", mesh[1]], cwd=case, stdout = gr, stderr = gr)
        elif suf[-1] == "neu":
            d = subprocess.Popen(["gambitToFoam", mesh[1]], cwd=case, stdout = gr, stderr = gr)
        elif suf[-1] == "msh":
            d = subprocess.Popen(["fluentMeshToFoam", mesh[1]], cwd=case, stdout = gr, stderr = gr)
        elif suf[-1] == "geo":
            d = subprocess.Popen(["cfx4ToFoam", mesh[1]], cwd=case, stdout = gr, stderr = gr)
        elif suf[-1] == "ans":
            d = subprocess.Popen(["ideasToFoam", mesh[1]], cwd=case, stdout = gr, stderr = gr)
        d.wait()
        gr.close()

    #Finds the boundaries from mesh
    def findBound(self):
        try:
            case
        except: 
            return
        bound = case+"/constant/polyMesh/boundary"
        f = open(bound, 'r+')
        lines = f.readlines()
        f.close()
        num = int(lines[17])
        first = []
        for i in range(1, num+1):
            j = 13+6*i
            first.append(lines[j][:-1])
        #Create boundaries in U,P,T etc
        try:
            properties
        except:
            return
        for f in properties:
            shutil.copy2(case+"/templates/"+f, case+"/templates/"+f+"_tmp")
        global con 
        con = len(first)
        for i in range(0, con):
            pb = first[i]+"\n    {\n        type            typ"+str(i)+"/*type*/;\n        value           val"+str(i)+"/*val*/; \n    }\n"
            for f in properties:
                textparse.parse(case+"/templates/"+f+"_tmp", pb, 23)
        #The next section creates the new "Boundaires" tab
        #Auto completer 
        c = QtGui.QCompleter(["fixedValue", "fixedGradient", "mixed", "calculated", "directionMixed", "zeroGradient", "movingWallVelocity", "pressureInletVelocity", "pressureDirectedInletVelocity", "surfaceNormalFixedValue", "totalPressure", "turbulentInlet", "buoyantPressure", "inletOutlet", "outletInlet", "pressureInletOutletVelocity", "pressureDirectedInletOutletVelocity", "pressureTransmissive", "supersonicFreeStream", "slip", "partialSlip", "patch", "symmetryPlane", "empty", "wedge", "cyclic", "wall", "processor"])
        c.setCompletionMode(QtGui.QCompleter.PopupCompletion)
        d = QtGui.QCompleter(['uniform (x y z)', 'uniform p'])
        d.setCompletionMode(QtGui.QCompleter.PopupCompletion)   
        newtab = QtGui.QScrollArea()
        newtab.setWidget(QtGui.QWidget())
        newtab_layout = QtGui.QVBoxLayout(newtab.widget())
        newtab.setWidgetResizable(True)
        global edit
        edit = {}
        #grid positioner
        grid = QtGui.QGridLayout()
        font = QtGui.QFont()
        font.setUnderline(True)
        font.setBold(True)
        #Add appropriate lables
        n = 1
        for f in properties:
            lable = QtGui.QLabel(str(f))
            lable.setFont(font)
            grid.addWidget(lable,0,n)
            n +=1
        font.setUnderline(False)
        j = 1
        i = 1
        while len(first) > 0:
            label = QtGui.QLabel(first.pop())
            label.setFont(font)
            grid.addWidget(label, j, 0)
            label = QtGui.QLabel("Type: ")
            grid.addWidget(label, j+1, 0)
            label = QtGui.QLabel("Value: ")
            grid.addWidget(label, j+2, 0)
            for x in range(1, n):
                edit[(i)] = QtGui.QComboBox()
                grid.addWidget(edit[(i)], j+1, x)
                edit[(i+1)] = QtGui.QComboBox()
                grid.addWidget(edit[(i+1)], j+2, x)
                edit[(i)].setEditable(True)
                edit[(i+1)].setEditable(True)
                edit[(i)].setCompleter(c)
                edit[(i+1)].setCompleter(d)
                i += 2
            j += 3
        newtab.setLayout(grid)
        #Create a new tab, then switch
        self.Schemes.addTab(newtab, "Boundary Conditions")
        self.Schemes.removeTab(5)

    #Sets the control in ControlDict
    def control(self):
        global run, delta, write, start
        cd = case+"/system/controlDict"
        start = self.textEdit_12.toPlainText()
        start = str(start)
        run = self.textEdit_4.toPlainText()
        run = str(run)
        delta = self.textEdit_5.toPlainText()
        delta = str(delta)
        write = self.textEdit_6.toPlainText()
        write = str(write)
        shutil.copy2(case+"/templates/controlDict", case+"/system/controlDict")
        replace.replace(cd, "STAART", start)
        replace.replace(cd, "TIMEEND", run)
        replace.replace(cd, "TDELTA", delta)
        replace.replace(cd, "INTERVALWRITE", write)

    #Finds the boundary conditions (groups from geometry)
    def cond(self): 
        i = 1
        for j in range(con-1, -1, -1):
            for f in properties:
                var = case+"/"+str(start)+"/"+str(f)
                typ = edit[(i)].currentText()
                val = edit[(i+1)].currentText()
                replace.replace(var, "typ"+str(j)+"/*type*/;", str(typ)+";")
                if typ == "zeroGradient":
                    replace.replace(var, "value           val"+str(j)+"/*val*/;", "")
                else:            
                    replace.replace(var, "val"+str(j)+"/*val*/;", str(val)+";")
                i += 2

    #Checks if leagal boundary condition type is given
    def condType(self):
        text = set()
        i = 1
        for j in range(0, con):
            ty1 = edit[(i)].currentText()
            ty1 = str(ty1)
            text.add(ty1)
            i += 2
        values = set(["fixedValue", "fixedGradient", "zeroGradient", "calculated", "mixed", "directionMixed",
                    "movingWallVelocity", "pressureInletVelocity", "pressureDirectedInletVelocity",
                    "surfaceNormalFixedValue", "totalPressure", "turbulentInlet", "fluxCorrectedVelocity",
                    "buoyantPressure", "inletOutlet", "outletInlet", "pressureInletOutletVelocity",
                    "pressureDirectedInletOutletVelocity", "pressureTransmissive", "supersonicFreeStream",
                    "slip", "partialSlip", "patch", "empty", "symmetryPlane", "wedge", "cyclic",
                    "wall", "processor"])
        if text.issubset(values):
                return True
        else:                
                return False

    #If turbulence is activated, get properties
    def turbulence(self, state):
        if state == QtCore.Qt.Unchecked:
            try:
                case
            except:
                self.textEdit.setEnabled(False)
                self.label_14.setEnabled(False)
                self.comboBox_4.setEnabled(False)
                self.Schemes.removeTab(7)
                self.turb = False
                return
            self.noturbulence()
            self.turb = False
            return 
        try:
            case
        except:
            QtGui.QMessageBox.question(self, 'An error has occured', "No case defined", QtGui.QMessageBox.Ok)
            self.checkBox.setCheckState(QtCore.Qt.Unchecked)
            return
        if state == QtCore.Qt.Checked:
            self.turb = True
            self.textEdit.setEnabled(True)
            self.label_14.setEnabled(True)
            self.comboBox_4.setEnabled(True)

    #Setup the turbulence tab
    def initturbulence(self):
        for f in turbprop:
            shutil.copy2(case+"/templates/"+f, case+"/templates/"+f+"_tmp")
        shutil.copy2(case+"/templates/turbulenceProperties", case+"/constant/turbulenceProperties")
        bound = case+"/constant/polyMesh/boundary"
        f = open(bound, 'r+')
        lines = f.readlines()
        f.close()
        num = int(lines[17])
        first = []
        for i in range(1, num+1):
            j = 13+6*i
            first.append(lines[j][:-1])
        global tcon
        tcon = len(first)
        for i in range(0, tcon):
            pb = first[i]+"\n    {\n        type            typ"+str(i)+"/*type*/;\n        value           val"+str(i)+"/*val*/; \n    }\n"
            for f in turbprop:
                textparse.parse(case+"/templates/"+f+"_tmp", pb, 23)           
        #Need to add options in tab
        newtab_2 = QtGui.QScrollArea()
        newtab_2.setWidget(QtGui.QWidget())
        newtab_layout = QtGui.QVBoxLayout(newtab_2.widget())
        newtab_2.setWidgetResizable(True)
        global redit
        redit = {}
        #Auto completer
        c = QtGui.QCompleter(["fixedValue", "zeroGradient"])
        c.setCompletionMode(QtGui.QCompleter.UnfilteredPopupCompletion)
        d = QtGui.QCompleter(['uniform (x y z)', 'uniform p'])
        d.setCompletionMode(QtGui.QCompleter.UnfilteredPopupCompletion)        
        #grid positioner
        grid = QtGui.QGridLayout()
        font = QtGui.QFont()
        font.setUnderline(True)
        font.setBold(True)
        n = 1
        for f in turbprop:
            lable = QtGui.QLabel(str(f))
            lable.setFont(font)
            grid.addWidget(lable,0,n)
            n += 1
        font.setUnderline(False)
        # Counters to keep control on boxes
        j = 1
        i = 1
        while len(first) > 0:
            label = QtGui.QLabel(first.pop())
            label.setFont(font)
            grid.addWidget(label, j, 0)
            label = QtGui.QLabel("Type: ")
            grid.addWidget(label, j+1, 0)
            label = QtGui.QLabel("Value: ")
            grid.addWidget(label, j+2, 0)
            for x in range(1, n):
                redit[(i)] = QtGui.QComboBox()
                grid.addWidget(redit[(i)], j+1, x)
                redit[(i+1)] = QtGui.QComboBox()
                grid.addWidget(redit[(i+1)], j+2, x)
                redit[(i)].setEditable(True)
                redit[(i+1)].setEditable(True)
                redit[(i)].setCompleter(c)
                redit[(i+1)].setCompleter(d)
                i += 2
            j += 3
        newtab_2.setLayout(grid)
        #Create a new tab, then switch
        self.Schemes.addTab(newtab_2, "Turbulence Properties")

    #Set the options nedded for turbulence
    def setturbulence(self):
        ras = case+"/constant/RASProperties"
        replace.replace(ras, "laminar;", str(self.comboBox_4.currentText())+";")
        replace.replace(ras, "off;", "on;") 
        tur = case+"/constant/turbulenceProperties"
        replace.replace(tur, "nu;", self.textEdit.toPlainText()+";")
        for f in turbprop:
            shutil.copy2(case+"/templates/"+f+"_tmp", case+"/"+str(start)+"/"+f)
        i = 1
        for j in range(con-1, -1, -1):
            for f in turbprop:
                var = case+"/"+str(start)+"/"+str(f)
                typ = redit[(i)].currentText()
                val = redit[(i+1)].currentText()
                replace.replace(var, "typ"+str(j)+"/*type*/;", str(typ)+";")
                if typ == "zeroGradient":
                    replace.replace(var, "value           val"+str(j)+"/*val*/;", "")
                else:            
                    replace.replace(var, "val"+str(j)+"/*val*/;", str(val)+";")
                i += 2

    #Remove turbulence related files and tabs
    def noturbulence(self):
        self.textEdit.setEnabled(False)
        self.label_14.setEnabled(False)
        self.comboBox_4.setEnabled(False)
        shutil.copy2(case+"/templates/RASProperties", case+"/constant/RASProperties")
        self.Schemes.removeTab(7)

    #Setting the viscosity
    def visc(self):
        shutil.copy2(case+"/templates/transportProperties", case+"/constant/transportProperties")
        vis = case+"/constant/transportProperties"
        replace.replace(vis, "nu;", self.textEdit_10.toPlainText()+";")

    #Allows for additional terminal commands
    def terminal(self):
        try:
            case
        except:
            QtGui.QMessageBox.question(self, 'An error has occured', "No case defined", QtGui.QMessageBox.Ok)
            return
        command = str(self.lineEdit_2.text())
        command = command.split(' ')
        self.lineEdit_2.clear()
        cmd = open(pwd+"/outp/terminal","w")
        if len(command) == 1:
            t = subprocess.Popen([command[0]], cwd=case, stdout = cmd, stderr = cmd)
        elif len(command) == 2:
            t = subprocess.Popen([command[0], command[1]], cwd=case, stdout = cmd, stderr = cmd)
        elif len(command) == 3:
            t = subprocess.Popen([command[0], command[1], command[2]], cwd=case, stdout = cmd, stderr = cmd)
        else:
            self.textEdit_9.append("Command not recognized")            
        try:
            t
        except:
            return
        t.wait()
        cmd.close()
        self.textEdit_9.append(open(pwd+"/outp/terminal").read())

    #Launches paraFOAM
    def para(self):
        try:
            case
        except:
            QtGui.QMessageBox.question(self, 'An error has occured', "No case defined", QtGui.QMessageBox.Ok)
            return
        e = subprocess.Popen(["paraFoam"], cwd = case)

    #Checks if input is a float/number 
    def validate(self):
        if not self.isFloat(self.textEdit.toPlainText(), "Viscosity is ill defined"):
            return False
        if not self.isFloat(self.textEdit_2.toPlainText(), "Pressure tolerace is ill defined"):
            return False
        if not self.isFloat(self.textEdit_3.toPlainText(), "Velocity  tolerace is ill defined"):
            return False
        if not self.isFloat(self.textEdit_4.toPlainText(), "Lenght of simulation is ill defined"):
            return False
        if not self.isFloat(self.textEdit_5.toPlainText(), "Timestep  is ill defined"):
            return False
        if not self.isFloat(self.textEdit_6.toPlainText(), "Write interval  is ill defined"):
            return False
        if not self.isFloat(self.textEdit_7.toPlainText(), "Pressure relTol is ill defined"):
            return False
        if not self.isFloat(self.textEdit_8.toPlainText(), "Velocity relTol is ill defined"):
            return False
        if not self.isFloat(self.textEdit_12.toPlainText(), "Starttime is ill defined"):
            return False

        return True

    #Checks if parameter is a float 
    def isFloat(self, inp, text):
        try: 
            float(inp)
            return True
        except ValueError:
            QtGui.QMessageBox.question(self, 'An error has occured', text, QtGui.QMessageBox.Ok)
            return False

if __name__ == '__main__':
    app = QtGui.QApplication(sys.argv)
    window = Main()
    window.show()
    sys.exit(app.exec_())
